/*
    This file is part of the KDE project, module kdesu.
    SPDX-FileCopyrightText: 1999, 2000 Geert Jansen <jansen@kde.org>

    SPDX-License-Identifier: GPL-2.0-only
*/

#ifndef KDESUSTUBPROCESS_P_H
#define KDESUSTUBPROCESS_P_H

#include "ptyprocess_p.h"

namespace KDESu
{
class StubProcessPrivate : public PtyProcessPrivate
{
};

}

#endif
